import os
from box import Box
from .PartDomesticator import GoldenGateDomesticator

this_dir = os.path.realpath(__file__)
standards_dir = os.path.join(os.path.dirname(this_dir), "assembly_standards")

BUILTIN_STANDARDS = Box({})
for fname in os.listdir(standards_dir):
    name, ext = os.path.splitext(fname)
    if ext != ".csv":
        continue
    path = os.path.join(standards_dir, fname)
    BUILTIN_STANDARDS[name] = GoldenGateDomesticator.standard_from_spreadsheet(
        path, name_prefix=name + "_"
    )



""" geneblocks/__init__.py """

# __all__ = []

from .PartDomesticator import PartDomesticator, GoldenGateDomesticator
from .reports import write_pdf_domestication_report
from .builtin_standards import BUILTIN_STANDARDS
from .batch_domestication import batch_domestication
from .biotools import (load_record, load_records, write_record,
                       random_dna_sequence)
from .BarcodesCollection import BarcodesCollection
from .version import __version__



from copy import deepcopy
import itertools

from Bio import SeqIO
import pandas
import proglog

import flametree
from sequenticon import sequenticon

from .PartDomesticator import PartDomesticator
from .reports import write_pdf_domestication_report
from .biotools import (
    sanitize_and_uniquify,
    sequence_to_record,
    annotate_record,
    write_record,
)


def detect_non_unique_elements(elements):
    seen = {}
    for e in elements:
        if e not in seen:
            seen[e] = [e]
        else:
            seen[e].append(e)
    return [(e, len(instances)) for e, instances in seen.items() if len(instances) > 1]


def batch_domestication(
    records,
    target,
    domesticator=None,
    standard=None,
    allow_edits=False,
    domesticated_suffix="",
    include_optimization_reports=True,
    include_original_records=True,
    barcodes=(),
    barcode_order="same_as_records",
    barcode_spacer="AA",
    logger="bar",
):
    """Domesticate a batch of parts according to some domesticator/standard.
    
    Examples
    --------

    >>> from genedom import BUILTIN_STANDARDS, batch_domestication
    >>> batch_domestication(some_records, standard=BUILTIN_STANDARDS.EMMA)


    Returns
    -------
      (n_fails, zip_data)
        Where n_fails indicate how many optimizations have failed, and
        zip_data is binary zip data (bytes), in the case where
        target="@memory".

    Parameters
    ----------

    records
      List of Bioython records to be domesticated.

    target
      Path to a folder, to a zip file, or "@memory" for in-memory report
      generation (the raw binary data of a zip archive is then returned).

    domesticator
      Either a single domesticator, to be used for all parts in the batch, or
      a function f(record) => appropriate_domesticator. Note that a "standard"
      can be provided instead.

    standard
      A StandardDomesticatorsSet object which will be used to attribute a
      specific domesticator to each part. See BUILTIN_STANDARDS for examples.

    allow_edits
      If False, sequences cannot be edited by the domesticator, only extended
      with flanks. If a sequence has for instance forbidden restriction sites,
      the domesticaton will fail for this sequence (and this will be noted in
      the report.

    domesticated_suffix
      Suffix to give to the domesticated parts names to differentiate them from
      the original parts (this is optional).

    include_optimization_reports
      If yes, some genbanks and pdfs will be produced to show how each part
      was domesticated. This is in particular informative when a domestication
      fails and you want to understand why.

    include_original_records
      Will include the input records into the final report folder/archive,
      for traceability.

    barcodes
      Either a list [(barcode_name, barcode),...] or a dictionary {name: bc} or
      a BarcodesCollection instance. If any of this is provided, the final
      parts will have a barcode added on the left (this barcode will be
      "outside" the part and won't appear in final constructs, but can be used
      to check that the part is the one you think if your samples get mixed up).
      Note that if there are less barcodes than parts, the barcodes will cycle
      and several parts may get the same barcode (which is generally fine).

    barcode_order
      Either "same_as_records", or "by_size" if you want your barcodes to be
      attributed from the smallest to the longest part in the batch.

    barcode_spacer
      Sequence to appear between the barcode and the left flank of the
      domesticated part.

    logger
      Either "bar" or None for no logger or any Proglog ProgressBarLogger.
    """
    non_unique_record_ids = detect_non_unique_elements([r.id for r in records])
    if len(non_unique_record_ids):
        raise ValueError(
            "The following record IDs have several occurences "
            "in the provided records, which would lead to "
            "overwritten record files: "
            + ", ".join(
                ["%s (%s)" % (e, instances) for (e, instances) in non_unique_record_ids]
            )
        )
    logger = proglog.default_bar_logger(logger, min_time_interval=0.2)
    root = flametree.file_tree(target, replace=True)
    domesticated_dir = root._dir("domesticated_genbanks")
    if include_original_records:
        original_dir = root._dir("original")
    if include_optimization_reports:
        errors_dir = root._dir("error_reports")
    if standard is not None:
        domesticator = standard.record_to_domesticator

    if hasattr(barcodes, "items"):
        barcodes = list(barcodes.items())
    if len(barcodes):
        barcodes = [b for b, r in zip(itertools.cycle(barcodes), records)]
    if barcode_order == "by_size":
        lengths = [len(r) for r in records]
        barcodes = [b for _, b in sorted(zip(lengths, barcodes))]

    infos = []

    domesticators = set()
    nfails = 0
    domesticated_records = []
    columns = [
        "Record",
        "Ordering Name",
        "Domesticator",
        "Domesticated Record",
        "Added bp",
        "Edited bp",
    ]

    # DOMESTICATE ALL PARTS, APPEND BARCODE, GATHER DATA
    for i, record in logger.iter_bar(record=list(enumerate(records))):
        record = deepcopy(record)
        original_id = record.id
        domesticated_id = record.id + domesticated_suffix
        domesticated_file_name = domesticated_id + ".gb"
        if isinstance(domesticator, PartDomesticator):
            record_domesticator = domesticator
        else:
            record_domesticator = domesticator(record)
        domesticators.add(record_domesticator)
        if include_optimization_reports:
            report_target = errors_dir._dir(record.id)
        else:
            report_target = None
        if len(barcodes):
            barcode = barcodes[i]
            if not isinstance(barcode, str):
                barcode_id, barcode = barcode
                barcode_id = " " + barcode_id
            else:
                barcode_id = ""
            barcode = sequence_to_record(barcode)
            annotate_record(barcode, label="BARCODE" + barcode_id)
        else:
            barcode = None
        # final, edits, report, success, msg
        domestication_results = record_domesticator.domesticate(
            record, report_target=report_target, edit=allow_edits
        )
        if not domestication_results.success:
            nfails += 1
        if barcode is not None:
            domestication_results.record_after = (
                barcode + barcode_spacer + domestication_results.record_after
            )
        domestication_results.record_after.original_id = original_id
        domestication_results.record_after.id = domesticated_id.replace(" ", "_")
        SeqIO.write(
            domestication_results.record_after,
            domesticated_dir._file(domesticated_file_name),
            "genbank",
        )
        domesticated_records.append(domestication_results.record_after)
        if include_original_records:
            write_record(
                record, original_dir._file(original_id + ".gb"),
            )
        n_edits = domestication_results.number_of_edits()
        added_bp = len(domestication_results.record_after) - len(record)
        before_seqicon = sequenticon(record, output_format="html_image")
        after_seqicon = sequenticon(
            domestication_results.record_after, output_format="html_image"
        )

        infos.append(
            {
                "id": original_id,
                "Record": before_seqicon + original_id,
                "Domesticator": record_domesticator.name,
                "Domesticated Record": ("Failed: " + domestication_results.message)
                if not domestication_results.success
                else (after_seqicon + domesticated_id),
                "Added bp": added_bp,
                "Edited bp": n_edits,
            }
        )
        if barcode is not None:
            infos[-1]["Barcode"] = barcode_id

    # WRITE PDF REPORT

    sanitizing_table = sanitize_and_uniquify([info["id"] for info in infos])
    order_id_dataframe = pandas.DataFrame(
        list(sanitizing_table.items()), columns=["sequence", "order_id"]
    )
    order_id_dataframe.to_csv(root._file("order_ids.csv").open("w"), index=False)
    for info in infos:
        info["Order ID"] = sanitizing_table[info["id"]]
    columns = [
        "Record",
        "Order ID",
        "Domesticator",
        "Domesticated Record",
        "Added bp",
        "Edited bp",
    ]
    if "Barcode" in infos[0]:
        columns.append("Barcode")
    infos_dataframe = pandas.DataFrame(infos, columns=columns)
    infos_dataframe.sort_values("Order ID", inplace=True)
    domesticators = sorted(domesticators, key=lambda d: d.name)
    write_pdf_domestication_report(
        root._file("Report.pdf"), infos_dataframe, domesticators
    )

    # WRITE THE SEQUENCES TO ORDER AS FASTA

    order_dir = root._dir("sequences_to_order", replace=True)
    for r in domesticated_records:
        r.id = sanitizing_table[r.original_id]
        r.name = ""
        r.description = ""
    SeqIO.write(domesticated_records, order_dir._file("sequences_to_order.fa"), "fasta")

    # WRITE THE SEQUENCES TO ORDER AS EXCEL

    df = pandas.DataFrame.from_records(
        sorted(
            [
                {
                    "sequence": str(rec.seq).upper(),
                    "length": len(rec),
                    "sequence name": rec.id,
                }
                for rec in domesticated_records
            ],
            key=lambda d: d["sequence name"],
        ),
        columns=["sequence name", "length", "sequence"],
    )
    df.to_excel(order_dir._file("sequences_to_order.xls").open("wb"), index=False)
    df.to_csv(order_dir._file("all_domesticated_parts.csv").open("w"), index=False)
    return nfails, root._close()



"""Defines central class BlockFinder."""
from collections import OrderedDict
import pandas
import numpy as np

from Bio import Restriction

from dnachisel import (
    AvoidPattern,
    annotate_record,
    Location,
    sequence_to_biopython_record,
    EnzymeSitePattern,
)

from ..StandardDomesticatorsSet import StandardDomesticatorsSet
from .PartDomesticator import PartDomesticator


def nan_to_empty_string(val):
    """Return the value unless it is NaN, then it returns an empty string."""
    return val if (isinstance(val, str) or not np.isnan(val)) else ""


class GoldenGateDomesticator(PartDomesticator):
    """Special domesticator class for Golden-Gate standards.

    Parameters
    ----------

    left_overhang
      4bp overhang to be added on the left.

    right_overhang
      4bp overhang to be added on the right.

    left_addition
      Extra sequence of DNA to be systematically added on the left of each part
      between the enzyme site and the rest of the sequence.

    right_addition
      Extra sequence to be systematically added on the right of each part
      between the enzyme site and the rest of the sequence.

    enzyme
      Enzyme used for the Golden Gate assembly. This enzyme will be added on
      the flanks of the sequence, and the internal sequence will be protected
      against sites from this enzyme during optimization.

    extra_avoided_sites
      Other enzymes from which the sequence should be protected during
      optimization in addition to the assembly ``enzyme``.

    description
      Description of the domesticator as it will appear in reports.

    name
      Name of the domesticator as it will appear in reports.

    constraints
      Either Dnachisel constraints or functions (sequence => constraint) to be
      applied to the sequence for optimization.

    objectives
      Either Dnachisel objectives or functions (sequence => objective) to be
      applied to the sequence for optimization.
    """

    def __init__(
        self,
        left_overhang,
        right_overhang,
        left_addition="",
        right_addition="",
        enzyme="BsmBI",
        extra_avoided_sites=(),
        description="Golden Gate domesticator",
        name="unnamed_domesticator",
        cds_by_default=False,
        constraints=(),
        objectives=(),
    ):
        self.enzyme = enzyme
        self.left_overhang = left_overhang
        left_overhang = sequence_to_biopython_record(left_overhang)
        self.right_overhang = right_overhang
        right_overhang = sequence_to_biopython_record(right_overhang)
        for seq in [left_overhang, right_overhang]:
            annotate_record(seq, label=str(seq.seq))
        enzyme_seq = Restriction.__dict__[enzyme].site
        enzyme_seq = sequence_to_biopython_record(enzyme_seq)
        annotate_record(enzyme_seq, label=enzyme)
        self.enzyme_seq = enzyme_seq
        left_flank = self.enzyme_seq + "A" + left_overhang + left_addition
        right_flank = (
            right_addition
            + right_overhang
            + (self.enzyme_seq + "A").reverse_complement()
        )
        self.extra_avoided_sites = extra_avoided_sites
        constraints = list(constraints) + [
            (
                lambda seq: AvoidPattern(
                    EnzymeSitePattern(enzyme),
                    location=Location(len(left_flank), len(left_flank) + len(seq)),
                )
            )
            for enz in ([enzyme] + list(extra_avoided_sites))
        ]
        PartDomesticator.__init__(
            self,
            left_flank=left_flank,
            right_flank=right_flank,
            constraints=constraints,
            objectives=objectives,
            description=description,
            name=name,
            cds_by_default=cds_by_default,
        )

    def __repr__(self):
        return "GgDomesticator[%s](%s-%s)" % (
            self.enzyme,
            self.left_overhang,
            self.right_overhang,
        )

    def __str__(self):
        return "GgDomesticator[%s](%s-%s)" % (
            self.enzyme,
            self.left_overhang,
            self.right_overhang,
        )

    def details_list(self):
        result = PartDomesticator.details_list(self) + [
            ("Enzyme", "%s (%s)" % (self.enzyme, str(self.enzyme_seq.seq))),
            ("Left overhang", self.left_overhang),
            ("Right overhang", self.right_overhang),
        ]
        extra_sites = self.extra_avoided_sites
        if len(extra_sites):
            result += [("Other avoided sites", ", ".join(extra_sites))]
        return result

    @staticmethod
    def standard_from_spreadsheet(path=None, dataframe=None, name_prefix=""):
        """Parse a spreadsheet into a standard with Golden Gate domesticators.

        The input should be a table with columns names as follows:
        slot_name, left_overhang, right_overhang, left_addition,
        right_addition, enzyme, extra_avoided_sites, description.

        Parameters
        ----------

        path
          Path to a CSV or XLS(X) file. A dataframe can be provided instead.

        dataframe
          A pandas Dataframe which can be provided instead of a path.
        """
        if path is not None:
            if path.lower().endswith(".csv"):
                dataframe = pandas.read_csv(path)
            else:
                dataframe = pandas.read_excel(path)
        return StandardDomesticatorsSet(
            OrderedDict(
                [
                    (
                        row.slot_name,
                        GoldenGateDomesticator(
                            left_overhang=row.left_overhang,
                            right_overhang=row.right_overhang,
                            left_addition=nan_to_empty_string(row.left_addition),
                            right_addition=nan_to_empty_string(row.right_addition),
                            enzyme=row.enzyme,
                            extra_avoided_sites=[
                                e.strip() for e in row.extra_avoided_sites.split(",")
                            ]
                            if hasattr(row.extra_avoided_sites, "split")
                            else [],
                            description=row.description,
                            cds_by_default=(row.is_cds == "yes")
                            if hasattr(row, "is_cds")
                            else False,
                            name=name_prefix + row.slot_name,
                        ),
                    )
                    for i, row in dataframe.iterrows()
                ]
            )
        )



import os
from dnachisel import (
    AllowPrimer,
    DnaOptimizationProblem,
    EnzymeSitePattern,
    EnforceSequence,
    EnforceGCContent,
    AvoidPattern,
    random_dna_sequence,
)
from collections import OrderedDict
from .biotools import sequence_to_record, annotate_record, write_record


class BarcodesCollection(OrderedDict):
    """Class representing a set of named barcode sequences.

    These barcodes are meant to be annealed with same-sequence primers for PCR
    or sequencing.

    The constructor taked a list [(name, barcode), ...] as an input.

    Use ``BarcodesCollection.from_specs(n_barcodes=25)`` to generate an
    instance with 25 compatible barcodes.
    """

    def __init__(self, barcodes):
        OrderedDict.__init__(self, barcodes)

    @staticmethod
    def from_specs(
        n_barcodes=96,
        barcode_length=20,
        spacer="AA",
        forbidden_enzymes=("BsaI", "BsmBI", "BbsI"),
        barcode_tmin=55,
        barcode_tmax=70,
        other_primer_sequences=(),
        heterodim_tmax=5,
        max_homology_length=10,
        include_spacers=True,
        names_template="B_%03d",
    ):
        """Return a BarcodesCollection object with compatible barcodes.

        Parameters
        ----------

        n_barcodes
          Number of barcodes to design.

        barcode_length
          Length of each barcode.

        spacer
          Spacer to place between each barcode during the optimization,
          ideally the same spacer that will be used when adding the barcode
          to a part.

        include_spacers
          Whether the spacers should be part of the final sequence of the
          barcodes (they still won't be considered part of the annealing
          primer and won't be used for melting temperature computations).

        forbidden_enzymes
          Name of enzymes whose sites should not be in the barcodes.

        barcode_tmin, barcode_tmax
          Interval of acceptable values for the melting temperature.

        other_primer_sequences
          External sequences with which the primers should not anneal.

        heterodim_tmax
          Max acceptable melting temperature for the annealing of a barcode
          and one of the other_primer_sequences.

        max_homology_length
          Maximal homology between any two barcodes in the sequence.

        names_template
          The template used to name barcode number "i".
        """
        unit_length = barcode_length + len(spacer)
        seq_len = n_barcodes * unit_length
        units_coordinates = [
            (i, i + unit_length) for i in range(0, seq_len, unit_length)
        ]
        constraints = [
            AvoidPattern(EnzymeSitePattern(enzyme)) for enzyme in forbidden_enzymes
        ]
        for start, end in units_coordinates:
            constraints += [
                AllowPrimer(
                    tmin=barcode_tmin,
                    tmax=barcode_tmax,
                    max_homology_length=max_homology_length,
                    avoid_heterodim_with=None,
                    max_heterodim_tm=5,
                    location=(start, end - len(spacer)),
                ),
                EnforceSequence(spacer, location=(end - len(spacer), end)),
                EnforceGCContent(
                    mini=0.4, maxi=0.6, location=(start, end - len(spacer))
                ),
            ]
        problem = DnaOptimizationProblem(
            sequence=random_dna_sequence(seq_len), constraints=constraints
        )
        problem.logger.ignored_bars.add("location")
        problem.resolve_constraints()

        barcodes = [problem.sequence[start:end] for (start, end) in units_coordinates]
        if not include_spacers:
            barcodes = [b[: -len(spacer)] for b in barcodes]
        names = [(names_template % (i + 1)) for i in range(len(barcodes))]
        return BarcodesCollection(zip(names, barcodes))

    def to_sequences_list(self):
        """Return a list of sequences ["ATTG...", "TTCTGT..."]"""
        return list(self.values())

    def to_fasta(self, path=None):
        """Return (and optionally write) a fasta string of the barcodes."""
        fasta = "\n\n".join(
            ["> %s\n%s" % (name, barcode) for name, barcode in self.items()]
        )
        if path is not None:
            with open(path, "w+") as f:
                f.write(fasta)
        else:
            return fasta

    def to_records(self, path=None):
        """Return (optionally write) individual Genbanks of the barcodes."""
        records = []
        for (name, barcode) in self.items():
            record = sequence_to_record(barcode)
            record.id = name
            annotate_record(record, label=name)
            records.append(record)
        if path is not None:
            for r in records:
                write_record(r, os.path.join(path, "%s.gb" % r.id))
        return records



from datetime import datetime
import os
import hashlib
from copy import deepcopy

from Bio import SeqIO
import matplotlib.pyplot as plt
import pandas
import jinja2
import weasyprint

import flametree
from pdf_reports import (write_report, pug_to_html, dataframe_to_html,
                         style_table_rows, add_css_class)
import pdf_reports

from .version import __version__

THIS_PATH = os.path.dirname(os.path.realpath(__file__))
ASSETS_PATH = os.path.join(THIS_PATH, "reports_assets")
DOMESTICATION_REPORT_TEMPLATE = os.path.join(ASSETS_PATH,
                                             "domestication_report.pug")
STYLESHEET = os.path.join(ASSETS_PATH, 'report_style.css')


def genedom_pug_to_html(template, **context):
    now = datetime.now().strftime("%Y-%m-%d %H:%M")
    defaults = {
        'genedom_sidebar_text': "Generated on %s by Genedom version %s" %
                                (now, __version__),
        'genedom_logo_url': os.path.join(ASSETS_PATH, 'imgs', 'logo.png'),
    }
    for k in defaults:
        if k not in context:
            context[k] = defaults[k]
    return pug_to_html(template, **context)


def write_pdf_domestication_report(target, domestication_infos, domesticators):
    summary_table = dataframe_to_html(
        domestication_infos,
        extra_classes=("definition",)
    )
    def tr_modifier(tr):
        tds = list(tr.find_all("td"))
        if len(tds) == 0:
            return
        if len(tds) == 6:
            name, order_id, dom, final, added, edited = tds
        else:
            name, order_id, dom, final, added, edited, barcode = tds
        if "Failed" in final.text:
            add_css_class(tr, "negative")
        if edited.text != "0":
            add_css_class(edited, "warning")
    summary_table = style_table_rows(summary_table, tr_modifier)
    html = genedom_pug_to_html(
        DOMESTICATION_REPORT_TEMPLATE,
        summary_table=summary_table,
        domesticators=domesticators
    )
    write_report(html, target, extra_stylesheets=(STYLESHEET,))



class DomesticationResult:
    """Class to contain and represent one result of a part domestication.

    Parameters
    ----------
    record_before
      Biopython record of the sequence before it is domesticated.

    record_after
      Biopython record of the sequence after it was domesticated.

    edits_record
      Biopython record annotated with every mutation introduced by the
      domestication.

    report_data
      Raw binary data of a PDF report.

    success
      Boolean indicating whether the domestication was possible or some
      constraints could not be satisfied.

    message
      String containing some information on the reason for failure.
    """

    def __init__(
        self, record_before, record_after, edits_record, report_data, success, message
    ):
        self.record_before = record_before
        self.record_after = record_after
        self.edits_record = edits_record
        self.report_data = report_data
        self.success = success
        self.message = message

    def summary(self):
        """Return a string summarizing how the domestication went.

        Either "SUCCESS - 42bp edited" or "FAILURE - (some help message)"
        """
        if self.success:
            return "SUCCESS - %d nucleotides edited." % self.number_of_edits()
        else:
            return "FAILURE - %s" % self.message

    def number_of_edits(self):
        return sum(
            [
                len(f)
                for f in self.edits_record.features
                if f.qualifiers.get("is_edit", False)
            ]
        )



"""Defines central class PartDomesticator."""

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

from dnachisel import (
    reverse_translate,
    DnaOptimizationProblem,
    EnforceTranslation,
    annotate_record,
    CodonOptimize,
    Location,
    sequence_to_biopython_record,
    AvoidChanges,
)

from dnachisel.reports import SpecAnnotationsTranslator
from ..DomesticationResult import DomesticationResult


class PartDomesticator:
    """Generic domesticator.

    Parameters
    ----------

    name
      Domesticator name as it will appear in reports etc.

    description
      Short domesticator description as it will appear in reports etc.

    left_flank
      String. Left addition to the sequence (homology arms, enzymes sites etc).

    right_flank
      String. Right addition to the sequence (homology arms, enz. sites etc).

    constraints
      Either Dnachisel constraints or functions (sequence => constraint) to be
      applied to the sequence for optimization.

    objectives
      Either Dnachisel objectives or functions (sequence => objective) to be
      applied to the sequence for optimization.

    simultaneous_mutations
      Number of sequences mutations to be applied simulatenously during
      optimization. A larger number creates more noise but could allow to
      solve tougher problems.

    minimize_edits
      If true, the optimizer will attempt to minimize changes while making
      sure the constraints hold (each edit incurs a penalty of 1 in the
      total optimization score).

    logger
      A proglog logger or 'bar' or None for no logger at all.
    """

    def __init__(
        self,
        name="unnamed domesticator",
        left_flank="",
        right_flank="",
        constraints=(),
        objectives=(),
        cds_by_default=False,
        description=None,
        simultaneous_mutations=1,
        minimize_edits=True,
        logger=None,
    ):
        if isinstance(left_flank, str):
            left_flank = sequence_to_biopython_record(left_flank)
            annotate_record(left_flank, label="left flank")
        if isinstance(right_flank, str):
            right_flank = sequence_to_biopython_record(right_flank)
            annotate_record(right_flank, label="right flank")
        self.name = name
        self.constraints = constraints
        self.left_flank = left_flank
        self.right_flank = right_flank
        self.constraints = list(constraints)
        self.objectives = list(objectives)
        self.description = description
        self.logger = logger
        self.simultaneous_mutations = simultaneous_mutations
        self.minimize_edits = minimize_edits
        self.cds_by_default = cds_by_default

    def domesticate(
        self,
        dna_sequence=None,
        protein_sequence=None,
        is_cds="default",
        codon_optimization=None,
        extra_constraints=(),
        extra_objectives=(),
        final_record_target=None,
        edit=False,
        barcode="",
        barcode_spacer="AA",
        report_target=None,
    ):
        """Domesticate a sequence.

        Parameters
        ----------

        dna_sequence
          The DNA sequence string to domesticate.

        protein_sequence
          Amino-acid sequence of the protein, which will be converted into
          a DNA sequence string.

        is_cds
          If True, sequence edits are restricted to synonymous mutations.

        codon_optimization
          Either None for no codon optimization or the name of an organism
          supported by DnaChisel.

        extra_constraints
          List of extra constraints to apply to the domesticated sequences.
          Each constraint is either a DnaChisel constraint or a function
          (dna_sequence => DnaChisel constraint).

        extra_objectives
          List of extra optimization objectives to apply to the domesticated
          sequences. Each objective is either a DnaChisel constraint or a
          function (dna_sequence => DnaChisel constraint).

        final_record_target
          Path to the file where to write the final genbank.

        edit
          Turn to True to allow sequence edits (if it is false and no all
          constraints are originally satisfied, a failed domestication result
          (i.e. with attribute ``success`` set to False) will be returned.

        report_target
          Target for the sequence optimization report (a folder path, or a zip
          path).

        barcode
          A sequence of DNA that will be added to the left of the sequence once
          the domestication is done.

        barcode_spacer
          Nucleotides to be added between the barcode and the enzyme (optional,
          the idea here is that they will make sure to avoid the creation of
          unwanted cutting sites).

        Returns
        -------

        final_record, edits_record, report_data, success, msg
        """
        if is_cds == "default":
            is_cds = self.cds_by_default
        if isinstance(dna_sequence, SeqRecord):
            problem = DnaOptimizationProblem.from_record(dna_sequence)
            for spec in problem.constraints + problem.objectives:
                spec.location += len(self.left_flank)
            extra_constraints = list(extra_constraints) + problem.constraints
            extra_objectives = list(extra_constraints) + problem.objectives

        if protein_sequence is not None:
            is_cds = True
            dna_sequence = reverse_translate(protein_sequence)
        constraints = [
            c(dna_sequence) if hasattr(c, "__call__") else c
            for c in list(extra_constraints) + self.constraints
        ]
        location = Location(
            len(self.left_flank), len(self.left_flank) + len(dna_sequence)
        )
        if is_cds:
            constraints.append(EnforceTranslation(location=location))
        objectives = [
            o(dna_sequence) if hasattr(o, "__call__") else o
            for o in list(extra_objectives) + self.objectives
        ]
        if codon_optimization:
            objectives.append(
                CodonOptimize(species=codon_optimization, location=location)
            )
        if self.minimize_edits:
            objectives.append(AvoidChanges())

        extended_sequence = self.left_flank + dna_sequence + self.right_flank

        if (not is_cds) and (not edit):
            constraints.append(AvoidChanges())
        problem = DnaOptimizationProblem(
            extended_sequence,
            constraints=constraints,
            objectives=objectives,
            logger=self.logger,
        )
        all_constraints_pass = problem.all_constraints_pass()
        no_objectives = (len(problem.objectives) - self.minimize_edits) == 0
        report_data = None
        optimization_successful = True
        message = ""
        # print (all_constraints_pass, no_objectives)
        if not (all_constraints_pass and no_objectives):
            problem.n_mutations = self.simultaneous_mutations

            if report_target is not None:
                (success, message, report_data) = problem.optimize_with_report(
                    target=report_target, project_name=self.name
                )
                optimization_successful = success
            else:
                report_data = None
                try:
                    problem.resolve_constraints()
                    problem.optimize()
                except Exception as err:
                    message = str(err)
                    optimization_successful = False
                    report_data = None
        final_record = problem.to_record(
            with_original_features=True,
            with_original_spec_features=False,
            with_constraints=False,
            with_objectives=False,
        )
        edits_record = problem.to_record(
            with_original_features=True,
            with_original_spec_features=False,
            with_constraints=False,
            with_objectives=False,
            with_sequence_edits=True,
        )
        if final_record_target is not None:
            SeqIO.write(final_record, final_record_target, "genbank")

        return DomesticationResult(
            problem.sequence_before,
            final_record,
            edits_record,
            report_data,
            optimization_successful,
            message,
        )

    def details_list(self):
        """List of details for representing the domesticator in reports."""
        return [
            (label, value)
            for (label, value) in [
                ("Name", self.name),
                ("Description", self.description),
                ("Left addition", str(self.left_flank.seq)),
                ("Right addition", str(self.right_flank.seq)),
            ]
            if value not in (None, "")
        ]

    def html_details(self):
        """HTML representation of the ``details_list``, for reports."""
        return "<br />".join(
            ["<b>%s</b>: %s" % (name, value) for (name, value) in self.details_list()]
        )

    @staticmethod
    def plot_record(record, ax=None):
        """Plot the given record with a custom DnaFeaturesViewer plotter."""
        translator = SpecAnnotationsTranslator()
        gr_record = translator.translate_record(record)
        return gr_record.plot(ax=ax)



from genedom import GoldenGateDomesticator, random_dna_sequence, write_record

sequence = random_dna_sequence(5000, seed=123)
domesticator = GoldenGateDomesticator("ATTC", "ATCG")
domestication_results = domesticator.domesticate(sequence, edit=True)
print (domestication_results.summary())
write_record(domestication_results.record_after, 'basic_example_after.gb')
write_record(domestication_results.edits_record, 'basic_example_edits.gb')


import os
from genedom import BUILTIN_STANDARDS, load_record, batch_domestication

records = [
    load_record(os.path.join('example_parts', filename), name=filename)
    for filename in os.listdir('example_parts')
]
batch_domestication(records, 'example_domestication_report.zip',
                    standard=BUILTIN_STANDARDS.EMMA)


class StandardDomesticatorsSet:
    def __init__(self, domesticators):
        self.domesticators = domesticators

    def list_overhangs(self):
        domesticators = list(self.domesticators.values())
        overhangs = [domesticators[0].left_overhang]
        for d in domesticators:
            for o in (d.right_overhang, d.left_overhang):
                if o not in overhangs:
                    overhangs.append(o)
        return overhangs

    def record_to_domesticator(self, record):
        return self.domesticators[record.id.split("_")[0]]



import ez_setup

ez_setup.use_setuptools()

from setuptools import setup, find_packages

exec(open("genedom/version.py").read())  # loads __version__

setup(
    name="genedom",
    version=__version__,
    author="Zulko",
    description="Genetic parts standardization",
    long_description=open("pypi-readme.rst").read(),
    license="MIT",
    url="https://github.com/Edinburgh-Genome-Foundry/genedom",
    keywords="genetic DNA part standardization synthetic biology",
    packages=find_packages(exclude="docs"),
    include_package_data=True,
    install_requires=(
        "dnachisel[reports]",
        "snapgene_reader",
        "pdf_reports",
        "pandas",
        "dna_features_viewer",
        "python-box",
        "openpyxl",
        "flametree",
        "sequenticon",
    ),
)



from genedom import BarcodesCollection

barcodes_collection = BarcodesCollection.from_specs(
    n_barcodes=96, barcode_length=20, spacer='AA',
    forbidden_enzymes=('BsaI', 'BsmBI', 'BbsI'),
    barcode_tmin=55, barcode_tmax=70,
    other_primer_sequences=(), heterodim_tmax=5,
    max_homology_length=10, include_spacers=True,
    names_template="B_%03d")

barcodes_collection.to_fasta('example_barcodes_collection.fa')


import os
import re
from copy import deepcopy
import numpy as np

from snapgene_reader import snapgene_file_to_seqrecord

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

try:
    # Biopython <1.78
    from Bio.Alphabet import DNAAlphabet

    has_dna_alphabet = True
except ImportError:
    # Biopython >=1.78
    has_dna_alphabet = False
from Bio.SeqFeature import SeqFeature, FeatureLocation
from Bio import SeqIO


complements_dict = {"A": "T", "T": "A", "C": "G", "G": "C"}


def random_dna_sequence(length, probas=None, seed=None):
    """Return a random DNA sequence ("ATGGCGT...") with the specified length.

    Parameters
    ----------

    length
      Length of the DNA sequence.

    proba
      Frequencies for the different nucleotides, for instance
      ``probas={"A":0.2, "T":0.3, "G":0.3, "C":0.2}``.
      If not specified, all nucleotides are equiprobable (p=0.25).

    seed
      The seed to feed to the random number generator. When a seed is provided
      the random results depend deterministically on the seed, thus enabling
      reproducibility.
    """
    if seed is not None:
        np.random.seed(seed)
    if probas is None:
        sequence = np.random.choice(list("ATCG"), length)
    else:
        bases, probas = zip(*probas.items())
        sequence = np.random.choice(bases, length, p=probas)
    return "".join(sequence)


formats_dict = {".fa": "fasta", ".gb": "genbank", ".gbk": "genbank", ".dna": "snapgene"}


def load_record(filename, linear=True, name="unnamed", capitalize=True):
    no_extension, extension = os.path.splitext(filename)
    fmt = formats_dict[extension]
    if fmt == "snapgene":
        record = snapgene_file_to_seqrecord(filename)
    else:
        record = SeqIO.read(filename, fmt)
    if capitalize:
        record.seq = record.seq.upper()
    record.linear = linear
    record.id = name
    record.name = name.replace(" ", "_")[:20]

    return record


def load_records(path, capitalize=True):
    if isinstance(path, (list, tuple)):
        return [record for p in path for record in load_records(p)]
    no_extension, extension = os.path.splitext(path)
    fmt = formats_dict[extension]
    if fmt == "snapgene":
        records = [snapgene_file_to_seqrecord(path)]
    else:
        records = list(SeqIO.parse(path, fmt))
    for i, record in enumerate(records):
        if capitalize:
            record.seq = record.seq.upper()
        if str(record.id) in ["None", "", "<unknown id>", ".", " "]:
            record.id = path.replace("/", "_").replace("\\", "_")
            if len(records) > 1:
                record.id += "_%04d" % i
    return records


def complement(sequence):
    return "".join(complements_dict[c] for c in sequence)


def reverse_complement(sequence):
    return complement(sequence)[::-1]


def sequence_to_record(sequence, features=()):
    if has_dna_alphabet:
        seq = Seq(sequence, alphabet=DNAAlphabet())
    else:
        seq = Seq(sequence)

    seqrecord = SeqRecord(seq, features=list(features))
    seqrecord.annotations["molecule_type"] = "DNA"

    return seqrecord


def annotate_record(
    seqrecord, location="full", feature_type="feature", margin=0, **qualifiers
):
    """Add a feature to a Biopython SeqRecord.

    Parameters
    ----------

    seqrecord
      The biopython seqrecord to be annotated.

    location
      Either (start, end) or (start, end, strand). (strand defaults to +1).

    feature_type
      The type associated with the feature.

    margin
      Number of extra bases added on each side of the given location.

    qualifiers
      Dictionary that will be the Biopython feature's `qualifiers` attribute.
    """

    if location == "full":
        location = (margin, len(seqrecord) - margin)

    strand = location[2] if len(location) == 3 else 1
    seqrecord.features.append(
        SeqFeature(
            FeatureLocation(location[0], location[1], strand),
            qualifiers=qualifiers,
            type=feature_type,
        )
    )


def sanitize_string(
    string, max_length=15, replacements=(("'", "p"), ("*", "s"), ("-", "_"))
):
    for old, new in replacements:
        string = string.replace(old, new)
    string = re.sub(r"[^a-zA-Z\d\S]", "_", string)
    return string[:max_length]


def sanitize_and_uniquify(
    strings, max_length=15, replacements=(("'", "p"), ("*", "s"), ("-", "_"))
):
    dejavu = set()
    table = {}
    for string in strings:
        newstring = sanitize_string(
            string, max_length=max_length, replacements=replacements
        )
        i = 1
        while newstring in dejavu:
            i += 1
            newstring = newstring[:-1] + str(i)
        dejavu.add(newstring)
        table[string] = newstring
    return table


def write_record(record, target, fmt="genbank"):
    """Write a record as genbank, fasta, etc. via Biopython, with fixes."""
    record = deepcopy(record)
    record.name = record.name[:20]
    if has_dna_alphabet:
        if str(record.seq.alphabet.__class__.__name__) != "DNAAlphabet":
            record.seq.alphabet = DNAAlphabet()
    record.annotations["molecule_type"] = "DNA"
    if hasattr(target, "open"):
        target = target.open("w")
    SeqIO.write(record, target, fmt)
